/*

  Gesture.java
  KanjiDraw

 Created by Julio Gorge on 07/11/2006.
 Copyright 2006-2007 LemonTeam.
 
 Kana DS is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Kana DS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Kana DS; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

import java.util.ArrayList;
import java.awt.geom.Point2D;
import java.awt.Graphics;
import java.awt.Graphics2D;

public class Gesture implements Cloneable
{
	boolean normalized = false;
	ArrayList strokes = null;
	
	
	// Clone object
	
	public Object clone()
	{
		Gesture gestureClone = new Gesture( strokes );		
		gestureClone.normalized = normalized;
		
		return (Object)gestureClone;
	}


	// Empty constructor

	public Gesture()
	{
		strokes = new ArrayList();
	}
	
	
	// Construct from a list of strokes
	
	public Gesture( ArrayList fromStrokes )
	{
		// "hard copy" everything
		
		strokes = new ArrayList();
		
		int strokeCount = fromStrokes.size();
		
		for ( int i = 0; i < strokeCount; i ++ )
		{
			Stroke stroke = (Stroke)fromStrokes.get( i );
			Stroke newStroke = new Stroke();
			
			int pointCount = stroke.size();
			
			for ( int j = 0; j < pointCount; j ++ )
			{
				Point2D.Float newPt = new Point2D.Float();
				Point2D.Float pt = (Point2D.Float)stroke.get( j );
				newPt.setLocation( pt.getX(), pt.getY() );
				newStroke.add( newPt );
			}
			
			strokes.add( newStroke );
		}
		
		normalized = false;
	}
	
	
	// Make the gesture empty: clear the stroke list
	
	public void clear()
	{
		if ( strokes != null )
			strokes.clear();
			
		normalized = false;
	}


	// Add a new stroke to the gesture
	
	public void addStroke( Stroke stroke )
	{
		strokes.add( stroke );
		normalized = false;
	}

	
	// Apply several normalization functions to the gesture's strokes

	public void normalize()
	{
		if ( !normalized )
		{
			int strokeCount = strokes.size();
			
			for ( int i = 0; i < strokeCount; i ++ )
			{
				Stroke stroke = (Stroke)strokes.get( i );
				strokes.set( i , (Stroke)stroke.normalizeSpacing() );
			}
			
			normalizeSize();
			//normalizeCenter();
			
			normalized = true;
			
			System.out.println( "gesture with " + strokeCount + " strokes normalized" );
		}
	}
	
	
	// Normalize size of the gesture's strokes
	
	private void normalizeSize()
	{
		float minX = Float.MAX_VALUE;
		float maxX = -Float.MAX_VALUE;
		float minY = Float.MAX_VALUE;
		float maxY = -Float.MAX_VALUE;

		//Calculate extents of the gesture
		
		int strokeCount = strokes.size();
		
		for ( int i = 0; i < strokeCount; i ++ )
		{
			Stroke stroke = (Stroke)strokes.get( i );
			
			int pointCount = stroke.size();
			
			for ( int j = 0; j < pointCount; j ++ )
			{
				Point2D.Float pt = (Point2D.Float)stroke.get( j );
				
				if ( minX > pt.getX() ) minX = (float)pt.getX();
				if ( minY > pt.getY() ) minY = (float)pt.getY();
				if ( maxX < pt.getX() ) maxX = (float)pt.getX();
				if ( maxY < pt.getY() ) maxY = (float)pt.getY();
			}
		}
		
	
		//Calculate dimensions 
		
		float width = maxX - minX;
		float height = maxY - minY;


		//find out scale
		float scale = (width > height) ? width:height;
		if ( scale <= 0.0f ) return; //Empty or a single point stroke!
		scale = 1.0f / scale;

		//Do the actual scaling
		
		for ( int i = 0; i < strokeCount; i ++ )
		{
			Stroke stroke = (Stroke)strokes.get( i );
			
			int pointCount = stroke.size();
			
			for ( int j = 0; j < pointCount; j ++ )
			{
				Point2D.Float pt = (Point2D.Float)stroke.get( j );
				pt.setLocation( pt.getX() * scale, pt.getY() * scale );
			}
		
		}
	}
	
	
	// Normalize center of the gesture's strokes
	
	private void normalizeCenter()
	{
		float centerX = 0.0f;
		float centerY = 0.0f;
		int pointCount = 0;
		
		//Calculate the centroid of the gesture
		int strokeCount = strokes.size();
		
		for ( int i = 0; i < strokeCount; i ++ )
		{
			Stroke stroke = (Stroke)strokes.get( i );

			pointCount += stroke.size();
			
			int strokeSize = stroke.size();
			//size should always be == kPointsPerStroke
			for ( int j = 0; j < strokeSize; j ++ )
			{
				Point2D.Float pt = (Point2D.Float)stroke.get( j );
				centerX += (float)pt.getX();
				centerY += (float)pt.getY();
			}
		}

		//Calculate centroid
		if ( pointCount <= 0 ) return; //empty gesture
		
		centerX /= pointCount;
		centerY /= pointCount;

		//To move the gesture into the origin, subtract centroid coordinates
		//from point coordinates
		
		for ( int i = 0; i < strokeCount; i ++ )
		{
			Stroke stroke = (Stroke)strokes.get( i );
			
			int strokeSize = stroke.size();
			
			for ( int j = 0; j < strokeSize; j ++ )
			{
				Point2D.Float pt = (Point2D.Float)stroke.get( j );
				pt.setLocation( pt.getX() - centerX, pt.getY() - centerY );
			}
		}
	}
	
	
	// Perform a dot-product comparing operation between two gestures
	
	public float dotProduct( Gesture operand )
	{
		if ( operand == null )
			return 0.f;
	
		if ( strokes.size() != operand.strokes.size() )
		{
			System.out.println( "matching FAILED: gestures don't have the same number of strokes" );
			return -1.f;
		}
		

		float dotProduct = 0.0f;
		
		int strokeCount = strokes.size();
		
		for ( int i = 0; i < strokeCount; i ++ )
		{
			Stroke stroke1 = (Stroke)strokes.get( i );
			Stroke stroke2 = (Stroke)operand.strokes.get( i );
			
			if ( stroke1.size() != stroke2.size() )
			{
				System.out.println( "matching FAILED: strokes have different number of points" );
				return -1.f;
			}
				
			int ptCount = stroke1.size();	// SHOULD BE kPointsBlah
			for ( int j = 0; j < ptCount; j ++ )
			{
				Point2D.Float pt1 = (Point2D.Float)stroke1.get( j );
				Point2D.Float pt2 = (Point2D.Float)stroke2.get( j );
				
				dotProduct += pt1.getX() * pt2.getX() + pt1.getY() * pt2.getY();
			}
		}

		return dotProduct;

	}
	
	
	// Compare two gestures and return a resemblance factor
	
	float match( Gesture gesture )
	{
		if ( gesture == null || strokes == null || gesture.strokes == null )
			return -1.f;
	
		if ( strokes.size() != gesture.strokes.size() )
		{
			System.out.println( "matching FAILED: gestures don't have the same number of strokes" );
			return -1.f;
		}
		
		normalize();
		
		gesture.normalize();
	
		float score = dotProduct( gesture );
	   
		if ( score <= 0.0f )
			return 0.0f; //No match for sure 
		
		//at this point our gesture-vectors are not quite normalized
		//   yet - their dot product with themselves is not 1.

	   //we normalize the score itself

	   //this is basically a version of a famous formula for a cosine of the 
	   //angle between 2 vectors:
	   //cos a = (u.v) / (sqrt(u.u) * sqrt(v.v)) = (u.v) / sqrt((u.u) * (v.v))
	   
	   score /= Math.sqrt( dotProduct( this ) * gesture.dotProduct( gesture ) );
	   return score;
	}
	
	
	// Draw a gesture using a Graphics2D interface
	
	void draw( Graphics2D g )
	{
		draw( g, 0, 0, 1.f, 1.f );
	}
	
	void draw( Graphics2D g, int x, int y, float scaleX, float scaleY )
	{
		int strokeCount = strokes.size();
		
		for ( int i = 0; i < strokeCount; i ++ )
		{
			Stroke stroke = (Stroke)strokes.get( i );
			
			stroke.draw( g, x, y, scaleX, scaleY );
		}
	}
	
}
